import dataclasses
import functools
import typing as tp
from abc import ABC, abstractmethod
from functools import partial

import jax.tree_util as jtu

from nnx import nodes, reprlib
from nnx.nn import initializers

A = tp.TypeVar("A")
B = tp.TypeVar("B")
C = tp.TypeVar("C", bound="Container[tp.Any]")
F = tp.TypeVar("F", bound=tp.Callable[..., tp.Any])
Sharding = tp.Tuple[str, ...]


class Container(ABC, tp.Generic[A]):
    __slots__ = ("_value",)

    def __init__(self, value: A):
        self._value = value

    @property
    def value(self) -> A:
        return self._value

    def replace_value(self: C, value: tp.Any) -> C:
        if isinstance(value, Container):
            if not self.is_equivalent(value):
                raise ValueError(
                    "Cannot replace value from incompatible container, "
                    f"expected {self}, got {value}"
                )
            value = value.value

        return self._replace_value(value)

    @abstractmethod
    def _replace_value(self: C, value: tp.Any) -> C:
        ...

    @abstractmethod
    def is_equivalent(self, other: tp.Any) -> bool:
        ...

    @abstractmethod
    def copy(self: C) -> C:
        ...


class Node(Container[A], reprlib.Representable):
    __slots__ = ()

    def __nnx_repr__(self):
        yield reprlib.Object(type=type(self))
        yield reprlib.Attr(
            "value", repr(self._value) if isinstance(self._value, str) else self._value
        )

    def __eq__(self, other: object) -> bool:
        return type(other) is Node and self._value == other._value

    def copy(self) -> "Node[A]":
        return Node(self._value)

    def _replace_value(
        self,
        value: B,
    ) -> "Node[B]":
        return Node(value)

    def is_equivalent(self, other: tp.Any) -> bool:
        return type(other) is Node


def _node_flatten(
    x: Node[tp.Any],
    *,
    with_keys: bool,
):
    if with_keys:
        node = (jtu.GetAttrKey("value"), x._value)
    else:
        node = x._value

    return (node,), None


def _node_unflatten(metadata: None, children: tp.Tuple[A]) -> Node[A]:
    return Node(children[0])


jtu.register_pytree_with_keys(
    Node,
    partial(_node_flatten, with_keys=True),
    _node_unflatten,
    flatten_func=partial(_node_flatten, with_keys=False),
)


@dataclasses.dataclass
class VarMetadata(tp.Generic[A]):
    value: A
    sharding: Sharding


def with_partitioning(
    initializer: initializers.Initializer,
    sharding: Sharding,
) -> initializers.Initializer:
    @functools.wraps(initializer)
    def wrapper(*args):
        return VarMetadata(initializer(*args), sharding)

    return wrapper  # type: ignore


def var_metadata(value: A, sharding: Sharding) -> VarMetadata[A]:
    return VarMetadata(value, sharding)


class Variable(Container[A], reprlib.Representable):
    __slots__ = ("_collection", "_sharding")

    def __init__(
        self,
        value: A,
        collection: str,
        sharding: tp.Optional[Sharding],
    ):
        if isinstance(value, VarMetadata):
            sharding = value.sharding if sharding is None else sharding
            value = tp.cast(A, value.value)

        self._value = value
        self._collection = collection
        self._sharding = sharding

    def __eq__(self, other: object) -> bool:
        return (
            type(other) is Variable
            and self._value == other._value
            and self._collection == other._collection
            and self._sharding == other._sharding
        )

    @property
    def collection(self) -> str:
        return self._collection

    @property
    def sharding(self) -> tp.Optional[Sharding]:
        return self._sharding

    def __nnx_repr__(self):
        yield reprlib.Object(type=type(self))
        yield reprlib.Attr("collection", repr(self._collection))
        yield reprlib.Attr(
            "value", repr(self._value) if isinstance(self._value, str) else self._value
        )
        if self._sharding is not None:
            yield reprlib.Attr("sharding", self._sharding)

    def copy(self) -> "Variable[A]":
        return Variable(self._value, self._collection, self._sharding)

    def _replace_value(
        self,
        value: B,
    ) -> "Variable[B]":
        return Variable(
            value,
            self._collection,
            self._sharding,
        )

    def is_equivalent(self, other: tp.Any) -> bool:
        return (
            type(other) is Variable
            and self._collection == other._collection
            and self._sharding == other._sharding
        )


def _variable_flatten(
    x: Variable[tp.Any],
    *,
    with_keys: bool,
):
    if with_keys:
        node = (jtu.GetAttrKey("value"), x._value)
    else:
        node = x._value

    return (node,), (x._collection, x._sharding)


def _variable_unflatten(
    metadata: tp.Tuple[str, tp.Optional[Sharding]], children: tp.Tuple[A]
) -> Variable[A]:
    return Variable(children[0], *metadata)


jtu.register_pytree_with_keys(
    Variable,
    partial(_variable_flatten, with_keys=True),
    _variable_unflatten,
    flatten_func=partial(_variable_flatten, with_keys=False),
)


class Static(Container[A], reprlib.Representable):
    def __hash__(self) -> int:
        return hash(self._value)

    def __eq__(self, other: tp.Any) -> bool:
        return type(other) is Static and self._value == other._value

    def copy(self) -> "Static[A]":
        return Static(self._value)

    def _replace_value(
        self,
        value: B,
    ) -> "Static[B]":
        return Static(value)

    def is_equivalent(self, other: tp.Any) -> bool:
        return type(other) is Static

    def __nnx_repr__(self):
        yield reprlib.Object(type=type(self))
        yield reprlib.Attr(
            "value", repr(self._value) if isinstance(self._value, str) else self._value
        )


def _static_flatten(x: Static[tp.Any]):
    return (), x._value


def _static_unflatten(metadata: A, _) -> Static[A]:
    return Static(metadata)


jtu.register_pytree_node(Static, _static_flatten, _static_unflatten)


# --------------------
# constructors
# --------------------


def check_container(f: F, *, num_arg: int) -> F:
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        args = list(args)
        value = args[num_arg]

        if isinstance(value, Container):
            container = value
            value = container.value
        else:
            container = None

        args[num_arg] = value
        output = f(*args, **kwargs)

        if container is not None and not container.is_equivalent(output):
            raise ValueError(f"Container {container} is not equivalent to {output}")

        return output

    return wrapper  # type: ignore


@partial(check_container, num_arg=1)
def var(
    collection: str,
    value: A,
    sharding: tp.Optional[Sharding] = None,
) -> A:
    return Variable(  # type: ignore
        value,
        collection=collection,
        sharding=sharding,
    )


def param(
    value: A,
    sharding: tp.Optional[Sharding] = None,
) -> A:
    return var("params", value, sharding=sharding)


@partial(check_container, num_arg=0)
def node(value: A) -> A:
    return Node(value)  # type: ignore


@partial(check_container, num_arg=0)
def static(value: A) -> A:
    return Static(value)  # type: ignore


# register nodes
nodes.register_node_type(Node)
nodes.register_node_type(Variable)



import contextlib
import dataclasses
import threading
import typing as tp
from abc import abstractmethod


@dataclasses.dataclass
class ReprContext(threading.local):
    indent_stack: tp.List[str] = dataclasses.field(default_factory=lambda: [""])


REPR_CONTEXT = ReprContext()


@dataclasses.dataclass
class Object:
    type: tp.Union[str, type]
    start: str = "("
    end: str = ")"
    value_sep: str = "="
    elem_indent: str = "  "
    empty_repr: str = ""


@dataclasses.dataclass
class Attr:
    key: str
    value: tp.Union[str, tp.Any]
    start: str = ""
    end: str = ""


class Representable:
    __slots__ = ()

    @abstractmethod
    def __nnx_repr__(self) -> tp.Iterator[tp.Union[Object, Attr]]:
        raise NotImplementedError

    def __repr__(self) -> str:
        return get_repr(self)


@contextlib.contextmanager
def add_indent(indent: str) -> tp.Iterator[None]:
    REPR_CONTEXT.indent_stack.append(REPR_CONTEXT.indent_stack[-1] + indent)

    try:
        yield
    finally:
        REPR_CONTEXT.indent_stack.pop()


def get_indent() -> str:
    return REPR_CONTEXT.indent_stack[-1]


def get_repr(obj: Representable) -> str:
    if not isinstance(obj, Representable):
        raise TypeError(f"Object {obj!r} is not representable")

    iterator = obj.__nnx_repr__()
    config = next(iterator)
    if not isinstance(config, Object):
        raise TypeError(f"First item must be Config, got {type(config).__name__}")

    def _repr_elem(elem: tp.Any) -> str:
        if not isinstance(elem, Attr):
            raise TypeError(f"Item must be Elem, got {type(elem).__name__}")

        value = elem.value if isinstance(elem.value, str) else repr(elem.value)

        if "\n" in value and not isinstance(elem.value, Representable):
            value = value.replace("\n", "\n" + get_indent())

        return (
            f"{get_indent()}{elem.start}{elem.key}{config.value_sep}{value}{elem.end}"
        )

    with add_indent(config.elem_indent):
        elems = list(map(_repr_elem, iterator))
    elems = ",\n".join(elems)

    if elems:
        elems = "\n" + elems + "\n" + get_indent()
    else:
        elems = config.empty_repr

    type_repr = config.type if isinstance(config.type, str) else config.type.__name__

    return f"{type_repr}{config.start}{elems}{config.end}"



import hashlib
import typing as tp
from types import MappingProxyType

import jax
import jax.tree_util as jtu

from nnx import errors, tracers

KeyArray = jax.Array
Counts = tp.Tuple[int, ...]


def _stable_hash(data: Counts) -> int:
    hash_str = " ".join(str(x) for x in data)
    _hash = hashlib.blake2s(hash_str.encode())
    hash_bytes = _hash.digest()
    # uint32 is represented as 4 bytes in big endian
    return int.from_bytes(hash_bytes[:4], byteorder="big")


class RngStream:
    __slots__ = ("_key", "_count", "_count_path", "_trace_state")

    def __init__(self, key: KeyArray, count: int = 0, count_path: Counts = ()):
        self._key = key
        self._count = count
        self._count_path = count_path
        self._trace_state = tracers.TraceState()

    @property
    def count(self) -> int:
        return self._count

    @property
    def count_path(self) -> Counts:
        return self._count_path

    def make_rng(self) -> jax.random.KeyArray:
        if not self._trace_state.is_valid():
            raise errors.TraceContextError(
                "Cannot use RngStream from a different trace level"
            )

        fold_data = _stable_hash(self._count_path + (self._count,))
        self._count += 1
        return jax.random.fold_in(self._key, fold_data)

    def fork(self) -> "RngStream":
        if not self._trace_state.is_valid():
            raise errors.TraceContextError(
                "Cannot use RngStream from a different trace level"
            )
        count_path = self._count_path + (self._count,)
        self._count += 1
        return RngStream(self._key, count_path=count_path)

    def split(self, n: int = 2) -> "RngStream":
        if not self._trace_state.is_valid():
            raise errors.TraceContextError(
                "Cannot use RngStream from a different trace level"
            )
        key = self.make_rng()
        key = jax.random.split(key, n)
        return RngStream(key)


def _rng_stream_flatten_with_keys(
    rng: RngStream,
) -> tp.Tuple[
    tp.Tuple[tp.Tuple[tp.Hashable, jax.random.KeyArray], ...],
    tp.Tuple[int, Counts],
]:
    return ((jtu.GetAttrKey("key"), rng._key),), (rng.count, rng.count_path)


def _rng_stream_unflatten(
    aux_data: tp.Tuple[int, Counts],
    children: tp.Tuple[jax.random.KeyArray, ...],
) -> RngStream:
    count, count_path = aux_data
    key = children[0]
    return RngStream(key, count, count_path)


def _rng_stream_flatten(rng: RngStream):
    return (rng._key,), (rng.count, rng.count_path)


jax.tree_util.register_pytree_with_keys(
    RngStream,
    _rng_stream_flatten_with_keys,
    _rng_stream_unflatten,
    flatten_func=_rng_stream_flatten,
)


class ContextDef:
    __slots__ = ("_rng_counts", "_flags")

    def __init__(
        self,
        rng_counts: tp.Tuple[tp.Tuple[str, Counts], ...],
        flags: tp.Tuple[tp.Tuple[str, bool], ...],
    ):
        self._rng_counts = rng_counts
        self._flags = flags

    def merge(self, keys: tp.Mapping[str, KeyArray]) -> "Context":
        rngs = {
            name: RngStream(keys[name], count=0, count_path=count_path)
            for name, count_path in self._rng_counts
        }
        return Context(rngs=rngs, flags=dict(self._flags))


class PureContext(tp.Tuple[tp.Dict[str, KeyArray], ContextDef]):
    @classmethod
    def new(cls, keys: tp.Dict[str, KeyArray], contextdef: ContextDef):
        return cls((keys, contextdef))

    @property
    def keys(self) -> tp.Dict[str, KeyArray]:
        return self[0]

    @property
    def contextdef(self) -> ContextDef:
        return self[1]

    def merge(self):
        return self.contextdef.merge(self.keys)


def _pure_context_flatten(pure_context: PureContext):
    return tuple(pure_context), None


def _pure_context_unflatten(
    aux_data: None,
    children: tp.Tuple[tp.Dict[str, RngStream], ContextDef],
) -> PureContext:
    return PureContext(children)


jtu.register_pytree_node(PureContext, _pure_context_flatten, _pure_context_unflatten)


class Context:
    __slots__ = ("_rngs", "_flags")

    def __init__(
        self,
        rngs: tp.Mapping[str, RngStream],
        flags: tp.Mapping[str, bool],
    ):
        self._rngs = rngs
        self._flags = flags

    def has_rng(self, name: str) -> bool:
        return name in self._rngs

    def make_rng(self, name: str) -> KeyArray:
        if name not in self._rngs:
            raise ValueError(f"Unknown Rng Stream: {name}")
        return self._rngs[name].make_rng()

    def copy(self) -> "Context":
        return Context(rngs=self._rngs, flags=self._flags)

    def has_flag(self, name: str) -> bool:
        return name in self._flags

    def get_flag(self, name: str) -> tp.Optional[bool]:
        return self._flags.get(name, None)

    def partition(self) -> PureContext:
        rngs = {name: stream.fork() for name, stream in self._rngs.items()}
        keys = {name: stream._key for name, stream in rngs.items()}
        rng_counts = tuple((name, stream.count_path) for name, stream in rngs.items())
        return PureContext.new(keys, ContextDef(rng_counts, tuple(self._flags.items())))


def context(
    params: tp.Union[int, KeyArray, RngStream, None] = None,
    *,
    flags: tp.Optional[tp.Mapping[str, bool]] = None,
    **rngs: tp.Union[int, KeyArray, RngStream],
) -> Context:
    _flags = flags or {}

    if params is not None:
        rngs["params"] = params

    _rngs = {
        name: RngStream(jax.random.PRNGKey(value))
        if isinstance(value, int)
        else RngStream(value)
        if isinstance(value, jax.Array)
        else value
        for name, value in rngs.items()
    }

    return Context(rngs=_rngs, flags=_flags)



import functools
import typing as tp

import jax
import jax.stages

from nnx import contextlib, partitioning, tracers
from nnx.module import Module, ModuleDef, PureModule
from nnx.state import State

A = tp.TypeVar("A")
F = tp.TypeVar("F", bound=tp.Callable[..., tp.Any])
G = tp.TypeVar("G", bound=tp.Callable[..., tp.Any])

AxisName = tp.Hashable
Leaf = tp.Any
Leaves = tp.List[Leaf]


class JitTransform(jax.stages.Wrapped):
    def __init__(
        self,
        fun: tp.Callable[..., tp.Any],
        stateful: bool,
        **jit_kwargs,
    ):
        @functools.partial(jax.jit, **jit_kwargs)
        def jitted_fn(pure_module: PureModule[Module], *args, **kwargs):
            if "ctx" in kwargs and isinstance(kwargs["ctx"], contextlib.PureContext):
                kwargs["ctx"] = kwargs["ctx"].merge()

            nnx_trace = tracers.get_top_trace((args, kwargs))
            with tracers.nnx_trace(nnx_trace):
                module = pure_module.merge()
                out = fun(module, *args, **kwargs)

            if self.stateful:
                updates = module.get_state()
                out = (updates, out)

            return out

        self.jitted_fn = jitted_fn
        self.stateful = stateful

    def __call__(self, module: tp.Any, *args, **kwargs):
        if not isinstance(module, Module):
            raise TypeError(f"Expected Module, got {type(module).__name__}")
        if "ctx" in kwargs and isinstance(kwargs["ctx"], contextlib.Context):
            kwargs["ctx"] = kwargs["ctx"].partition()

        pure_module = module.partition()
        out = self.jitted_fn(pure_module, *args, **kwargs)
        if self.stateful:
            updates: State
            updates, out = out
            module.update_state(updates)
        return out

    def __repr__(self):
        return f"JitTransform({self.jitted_fn})"

    def lower(self, *args, **kwargs):
        return self.jitted_fn.lower(*args, **kwargs)


UNSPECIFIED = object()


def jit(
    fun: tp.Callable[..., tp.Any],
    *,
    stateful: bool = True,
    in_shardings: tp.Any = UNSPECIFIED,
    out_shardings: tp.Any = UNSPECIFIED,
    static_argnums: tp.Union[int, tp.Sequence[int], None] = None,
    static_argnames: tp.Union[str, tp.Iterable[str], None] = None,
    donate_argnums: tp.Union[int, tp.Sequence[int]] = (),
    keep_unused: bool = False,
    device: tp.Optional[jax.Device] = None,
    backend: tp.Optional[str] = None,
    inline: bool = False,
    abstracted_axes: tp.Optional[tp.Any] = None,
) -> jax.stages.Wrapped:
    if static_argnames is None:
        static_argnames = []
    elif isinstance(static_argnames, str):
        static_argnames = [static_argnames]
    else:
        static_argnames = list(static_argnames)

    static_argnames.append("_nnx__dagdef")

    jit_kwargs = dict(
        static_argnums=static_argnums,
        static_argnames=static_argnames,
        donate_argnums=donate_argnums,
        keep_unused=keep_unused,
        device=device,
        backend=backend,
        inline=inline,
        abstracted_axes=abstracted_axes,
    )

    if in_shardings is not UNSPECIFIED:
        jit_kwargs["in_shardings"] = in_shardings
    if out_shardings is not UNSPECIFIED:
        jit_kwargs["out_shardings"] = out_shardings

    ref_jit = JitTransform(
        fun,
        stateful,
        **jit_kwargs,
    )
    ref_jit = functools.wraps(fun)(ref_jit)
    # _update_decorator_fields(ref_jit, fun)
    return ref_jit


class GradTransform:
    def __init__(
        self,
        fun: tp.Callable[..., tp.Any],
        stateful: bool,
        predicate: partitioning.Predicate,
        has_aux: bool,
        holomorphic: bool,
        allow_int: bool,
        reduce_axes: tp.Sequence[AxisName],
    ):
        @functools.partial(
            jax.grad,
            argnums=0,  # we'll handle this ourselves
            has_aux=has_aux or stateful,
            holomorphic=holomorphic,
            allow_int=allow_int,
            reduce_axes=reduce_axes,
        )
        def grad_fn(
            diff: State,
            non_diff: State,
            moduledef: ModuleDef[Module],
            *args: tp.Any,
        ):
            with tracers.nnx_trace(tracers.get_top_trace(diff)):
                module = moduledef.merge(diff, non_diff)
                out = fun(module, *args)

            if self.stateful:
                updates = module.get_state()
                if self.has_aux:
                    loss, aux = out
                    out = (loss, (updates, aux))
                else:
                    out = (out, updates)

            return out

        self.grad_fn = grad_fn
        self.predicate = predicate
        self.has_aux = has_aux
        self.stateful = stateful

    def __call__(self, module: Module, *args: tp.Any):
        if not isinstance(module, Module):
            raise TypeError(f"Expected a Module, got {type(module).__name__}")

        (diff, nondiff), moduledef = module.partition(self.predicate, ...)

        grads = self.grad_fn(diff, nondiff, moduledef, *args)

        if self.stateful:
            updates: State
            if self.has_aux:
                grads, (updates, aux) = grads
                out = grads, aux
            else:
                out, updates = grads
            module.update_state(updates)
        else:
            out = grads

        return out

    def __repr__(self):
        return f"GradTransform({self.grad_fn})"


@tp.overload
def grad(
    fun: tp.Callable[..., tp.Any],
    wrt: partitioning.CollectionFilter = "params",
    *,
    stateful: bool = True,
    holomorphic: bool = False,
    allow_int: bool = False,
    reduce_axes: tp.Sequence[AxisName] = (),
) -> tp.Callable[..., State]:
    ...


@tp.overload
def grad(
    fun: tp.Callable[..., tp.Any],
    wrt: partitioning.CollectionFilter = "params",
    *,
    stateful: bool = True,
    has_aux: tp.Literal[True],
    holomorphic: bool = False,
    allow_int: bool = False,
    reduce_axes: tp.Sequence[AxisName] = (),
) -> tp.Callable[..., tp.Tuple[State, tp.Any]]:
    ...


def grad(
    fun: tp.Callable[..., tp.Any],
    wrt: partitioning.CollectionFilter = "params",
    *,
    stateful: bool = True,
    has_aux: bool = False,
    holomorphic: bool = False,
    allow_int: bool = False,
    reduce_axes: tp.Sequence[AxisName] = (),
) -> tp.Callable[..., tp.Union[tp.Tuple[State, tp.Any], State]]:
    predicate = partitioning.to_predicate(wrt)
    ref_grad = GradTransform(
        fun,
        stateful=stateful,
        predicate=predicate,
        has_aux=has_aux,
        holomorphic=holomorphic,
        allow_int=allow_int,
        reduce_axes=reduce_axes,
    )
    ref_grad = functools.wraps(fun)(ref_grad)
    # _update_decorator_fields(ref_grad, fun)
    return ref_grad



import contextlib
import dataclasses
import importlib.util
import inspect
import typing as tp
from abc import ABCMeta
from copy import copy
from functools import partial
from types import MappingProxyType

import jax

from nnx import nodes
from nnx.containers import Container

A = tp.TypeVar("A")
P = tp.TypeVar("P", bound="Pytree")


@contextlib.contextmanager
def _mutable(obj: P) -> tp.Iterator[None]:
    vars(obj)["_pytree__is_mutable"] = True
    try:
        yield
    finally:
        del vars(obj)["_pytree__is_mutable"]


@contextlib.contextmanager
def _initializing(obj: P) -> tp.Iterator[None]:
    vars(obj)["_pytree__initializing"] = True
    try:
        yield
    finally:
        del vars(obj)["_pytree__initializing"]


class PytreeMeta(ABCMeta):
    if not tp.TYPE_CHECKING:

        def __call__(cls: tp.Type[P], *args: tp.Any, **kwargs: tp.Any) -> P:
            return cls.call(*args, **kwargs)

    def call(cls: tp.Type[P], *args: tp.Any, **kwargs: tp.Any) -> P:
        obj: P = cls.__new__(cls, *args, **kwargs)
        vars(obj)["_pytree__sorted_fields"] = ["_pytree__sorted_fields"]

        with _mutable(obj), _initializing(obj):
            obj.__init__(*args, **kwargs)

        if dataclasses.is_dataclass(obj):
            assert isinstance(obj, Pytree)
            for field in dataclasses.fields(obj):
                if "nnx_container_fn" not in field.metadata:
                    continue

                container_fn = field.metadata["nnx_container_fn"]
                value = vars(obj)[field.name]
                value = container_fn(value)
                vars(obj)[field.name] = value

        vars(obj)["_pytree__sorted_fields"] = sorted(vars(obj))

        return obj


class Pytree(metaclass=PytreeMeta):
    _pytree__is_mutable: bool
    _pytree__class_is_mutable: bool
    _pytree__sorted_fields: tp.Tuple[str, ...]

    if not tp.TYPE_CHECKING:

        def __getattribute__(self, name: str) -> tp.Any:
            value = object.__getattribute__(self, name)
            if isinstance(value, Container):
                return value.value
            return value

        def __setattr__(self, name: str, value: tp.Any) -> None:
            self._setattr(name, value)

    def _setattr(self: P, name: str, value: tp.Any):
        vars_dict = vars(self)
        if "_pytree__initializing" in vars_dict:
            pass
        elif name not in vars_dict:
            raise AttributeError(r"Cannot add new fields to an initialized Pytree")
        elif (
            "_pytree__is_mutable" not in vars_dict
            and not self._pytree__class_is_mutable
        ):
            raise AttributeError(
                f"{type(self)} is immutable, trying to update field {name}"
            )

        if name in vars_dict and isinstance(vars_dict[name], Container):
            vars_dict[name] = vars_dict[name].replace_value(value)
        else:
            if isinstance(value, Container):
                value = value.copy()
            vars_dict[name] = value

    def __init_subclass__(cls, mutable: bool = False):
        super().__init_subclass__()
        # init class variables
        cls._pytree__is_mutable = False
        cls._pytree__class_is_mutable = mutable

        # TODO: clean up this in the future once minimal supported version is 0.4.7
        if hasattr(jax.tree_util, "register_pytree_with_keys"):
            if (
                "flatten_func"
                in inspect.signature(jax.tree_util.register_pytree_with_keys).parameters
            ):
                jax.tree_util.register_pytree_with_keys(
                    cls,
                    partial(
                        cls._pytree__flatten,
                        with_key_paths=True,
                    ),
                    cls._pytree__unflatten,
                    flatten_func=partial(
                        cls._pytree__flatten,
                        with_key_paths=False,
                    ),
                )
            else:
                jax.tree_util.register_pytree_with_keys(
                    cls,
                    partial(
                        cls._pytree__flatten,
                        with_key_paths=True,
                    ),
                    cls._pytree__unflatten,
                )
        else:
            jax.tree_util.register_pytree_node(
                cls,
                partial(
                    cls._pytree__flatten,
                    with_key_paths=False,
                ),
                cls._pytree__unflatten,
            )

        # flax serialization support
        if importlib.util.find_spec("flax") is not None:
            from flax import serialization

            serialization.register_serialization_state(
                cls, cls._to_flax_state_dict, cls._from_flax_state_dict
            )

    @classmethod
    def _pytree__flatten(
        cls,
        pytree: "Pytree",
        *,
        with_key_paths: bool,
    ):
        all_vars = vars(pytree)
        static = {}
        node_values = []
        node_names = []

        for field in pytree._pytree__sorted_fields:
            value = all_vars[field]

            if nodes.is_node(value):
                node_names.append(field)
                if with_key_paths:
                    node_values.append((jax.tree_util.GetAttrKey(field), value))
                else:
                    node_values.append(value)
            else:
                static[field] = value

        return node_values, (tuple(node_names), MappingProxyType(static))

    @classmethod
    def _pytree__unflatten(
        cls: tp.Type[P],
        metadata: tp.Tuple[tp.Tuple[str, ...], tp.Mapping[str, tp.Any]],
        node_values: tp.Tuple[tp.Any, ...],
    ) -> P:
        node_names, static_fields = metadata
        pytree = object.__new__(cls)
        pytree.__dict__.update(zip(node_names, node_values))
        pytree.__dict__.update(static_fields)
        return pytree

    @classmethod
    def _to_flax_state_dict(cls, pytree: "Pytree") -> tp.Dict[str, tp.Any]:
        from flax import serialization

        state_dict = {
            name: serialization.to_state_dict(getattr(pytree, name))
            for name, value in vars(pytree).items()
            if nodes.is_node(value)
        }
        return state_dict

    @classmethod
    def _from_flax_state_dict(
        cls,
        pytree: P,
        state: tp.Dict[str, tp.Any],
    ) -> P:
        """Restore the state of a data class."""
        from flax import serialization

        state = state.copy()  # copy the state so we can pop the restored fields.
        updates = {}
        for name, value in vars(pytree).items():
            if not nodes.is_node(value):
                continue
            if name not in state:
                raise ValueError(
                    f"Missing field {name} in state dict while restoring"
                    f" an instance of {type(pytree).__name__},"
                    f" at path {serialization.current_path()}"
                )
            value_state = state.pop(name)
            updates[name] = serialization.from_state_dict(value, value_state, name=name)
        if state:
            names = ",".join(state.keys())
            raise ValueError(
                f'Unknown field(s) "{names}" in state dict while'
                f" restoring an instance of {type(pytree).__name__}"
                f" at path {serialization.current_path()}"
            )
        return pytree.replace(**updates)

    def replace(self: P, **kwargs: tp.Any) -> P:
        """
        Replace the values of the fields of the object with the values of the
        keyword arguments. If the object is a dataclass, `dataclasses.replace`
        will be used. Otherwise, a new object will be created with the same
        type as the original object.
        """
        if dataclasses.is_dataclass(self):
            return dataclasses.replace(self, **kwargs)

        unknown_keys = set(kwargs) - set(vars(self))
        if unknown_keys and not self._pytree__class_is_mutable:
            raise ValueError(
                f"Trying to replace unknown fields {unknown_keys} "
                f"for '{type(self).__name__}'"
            )

        pytree = copy(self)
        with _mutable(pytree):
            for key, value in kwargs.items():
                setattr(pytree, key, value)

        return pytree


# register node types
nodes.register_node_type(Pytree)



import typing as tp

import jax
import jax.tree_util as jtu

from nnx import nodes, partitioning, reprlib
from nnx.containers import Variable

A = tp.TypeVar("A")

Leaf = tp.Any
Path = tp.Tuple[str, ...]
StateDict = tp.Dict[Path, tp.Any]
StateMapping = tp.Mapping[Path, tp.Any]


class State(tp.Mapping[tp.Tuple[str, ...], Leaf], reprlib.Representable):
    __slots__ = ("_mapping",)

    def __init__(
        self,
        __input: tp.Union[
            tp.Mapping[tp.Tuple[str, ...], Leaf],
            tp.Iterator[tp.Tuple[tp.Tuple[str, ...], Leaf]],
        ],
        /,
    ):
        if isinstance(__input, tp.Mapping):
            self._mapping = dict(sorted(__input.items(), key=lambda x: x[0]))
        else:
            self._mapping = dict(sorted(__input, key=lambda x: x[0]))

    def get_collections(self) -> tp.Set[tp.Union[str, None]]:
        return {
            value.collection if isinstance(value, Variable) else None
            for value in self._mapping.values()
        }

    def __getitem__(self, __key: tp.Tuple[str, ...]) -> Leaf:
        return self._mapping[__key]

    def __iter__(self) -> tp.Iterator[tp.Tuple[str, ...]]:
        return iter(self._mapping)

    def __len__(self) -> int:
        return len(self._mapping)

    def __nnx_repr__(self):
        yield reprlib.Object(type(self), value_sep=": ", start="({", end="})")

        for k, v in self._mapping.items():
            yield reprlib.Attr(str(k), v)

    @tp.overload
    def partition(self, first: partitioning.CollectionFilter, /) -> "State":
        ...

    @tp.overload
    def partition(
        self,
        first: partitioning.CollectionFilter,
        second: partitioning.CollectionFilter,
        /,
        *filters: partitioning.CollectionFilter,
    ) -> tp.Tuple["State", ...]:
        ...

    def partition(
        self,
        first: partitioning.CollectionFilter,
        /,
        *filters: partitioning.CollectionFilter,
    ) -> tp.Union["State", tp.Tuple["State", ...]]:
        *states, rest = _split_state(self, first, *filters)

        if rest:
            raise ValueError(
                f"Non-exhaustive filters, got a non-empty remainder: "
                f"{list(rest.keys())}.\nUse `...` to match all remaining elements."
            )

        if len(states) == 1:
            states = State(states[0])
        else:
            states = tuple(State(state) for state in states)
        return states

    @tp.overload
    def filter(
        self,
        first: partitioning.CollectionFilter,
        /,
    ) -> "State":
        ...

    @tp.overload
    def filter(
        self,
        first: partitioning.CollectionFilter,
        second: partitioning.CollectionFilter,
        /,
        *filters: partitioning.CollectionFilter,
    ) -> tp.Tuple["State", ...]:
        ...

    def filter(
        self,
        first: partitioning.CollectionFilter,
        /,
        *filters: partitioning.CollectionFilter,
    ) -> tp.Union["State", tp.Tuple["State", ...]]:
        (*states, _rest) = _split_state(self, first, *filters)

        assert len(states) == len(filters) + 1

        if len(states) == 1:
            states = State(states[0])
        else:
            states = tuple(State(state) for state in states)

        return states

    @staticmethod
    def merge(state: "State", /, *states: "State") -> "State":
        states = (state, *states)

        if len(states) == 1:
            return states[0]

        new_state: StateDict = {}

        for state in states:
            new_state.update(state)

        return State(new_state)

    def __or__(self, other: "State") -> "State":
        if not other:
            return self
        return State.merge(self, other)

    def __sub__(self, other: "State") -> "State":
        if not other:
            return self

        # create new State via __new__ to avoid __init__ sorting
        _mapping = {k: v for k, v in self.items() if k not in other}
        state = object.__new__(State)
        state._mapping = _mapping
        return state


def _state_flatten_with_keys(
    x: State,
):
    children = tuple((jtu.DictKey(key), value) for key, value in x.items())
    return children, tuple(x.keys())


def _state_unflatten(
    keys: tp.Tuple[Path, ...],
    leaves: tp.Tuple[Leaf, ...],
):
    state = object.__new__(State)
    state._mapping = dict(zip(keys, leaves))
    return state


jax.tree_util.register_pytree_with_keys(
    State, _state_flatten_with_keys, _state_unflatten
)


def _split_state(
    state: StateMapping,
    *filters: partitioning.CollectionFilter,
) -> tp.Tuple[StateDict, ...]:
    predicates = tuple(map(partitioning.to_predicate, filters))

    # we have n + 1 states, where n is the number of predicates
    # the last state is for values that don't match any predicate
    states: tp.Tuple[StateDict, ...] = tuple({} for _ in range(len(predicates) + 1))

    for path, value in state.items():
        for i, predicate in enumerate(predicates):
            if predicate(path, value):
                states[i][path] = value
                break
        else:
            # if we didn't break, set leaf to last state
            states[-1][path] = value

    return states


# register nodes
nodes.register_node_type(State)



__version__ = "0.0.0"


from .containers import (
    Container,
    Node,
    Static,
    Variable,
    VarMetadata,
    node,
    param,
    static,
    var,
    var_metadata,
    with_partitioning,
)
from .contextlib import Context, RngStream, context
from .dataclasses import (
    dataclass,
    field,
    node_field,
    param_field,
    static_field,
    var_field,
)
from .errors import TraceContextError
from .helpers import Map, Sequence, TrainState
from .module import Module, ModuleDef, PureModule
from .nn import initializers
from .nn.activations import (
    celu,
    elu,
    gelu,
    glu,
    hard_sigmoid,
    hard_silu,
    hard_swish,
    hard_tanh,
    leaky_relu,
    log_sigmoid,
    log_softmax,
    logsumexp,
    normalize,
    one_hot,
    relu,
    relu6,
    selu,
    sigmoid,
    silu,
    soft_sign,
    softmax,
    softplus,
    standardize,
    swish,
    tanh,
)
from .nn.linear import Conv, Embed, Linear
from .nn.normalization import BatchNorm, LayerNorm
from .nn.stochastic import Dropout
from .nodes import is_node, register_node_type
from .partitioning import buffers
from .pytreelib import Pytree
from .state import State
from .transforms import grad, jit



import typing as tp

import optax

from nnx import containers, pytreelib, utils
from nnx.contextlib import Context
from nnx.module import ApplyCaller, Module, PureModule
from nnx.state import State

A = tp.TypeVar("A")
M = tp.TypeVar("M", bound=Module)


class Map(Module, tp.Mapping[str, A]):
    @tp.overload
    def __init__(self, __iterable: tp.Iterable[tp.Tuple[str, A]]):
        ...

    @tp.overload
    def __init__(self, __mapping: tp.Optional[tp.Mapping[str, A]] = None, **kwargs: A):
        ...

    def __init__(self, *args, **kwargs):
        for name, value in dict(*args, **kwargs).items():
            setattr(self, name, value)

    def __getitem__(self, key) -> A:
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getattr__(self, key) -> A:
        return super().__getattribute__(key)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)

    def __iter__(self) -> tp.Iterator[str]:
        return iter(vars(self))

    def __len__(self) -> int:
        return len(vars(self))


class Sequence(Module, tp.Generic[A]):
    def __init__(self, iterable: tp.Iterable[A]):
        i = 0
        for i, value in enumerate(iterable):
            setattr(self, str(i), value)
        self._length = i + 1

    def __getitem__(self, key: int) -> A:
        if key >= len(self):
            raise IndexError(f"index {key} out of range for {self}")
        return getattr(self, str(key))

    def __iter__(self) -> tp.Iterator[A]:
        for i in range(len(self)):
            yield getattr(self, str(i))

    def __len__(self) -> int:
        return self._length

    def __call__(self, *args, ctx: tp.Optional[Context] = None, **kwargs) -> tp.Any:
        output: tp.Any = None

        for i, f in enumerate(self):
            if not callable(f):
                raise TypeError(f"Sequence[{i}] is not callable: {f}")
            if i > 0:
                if isinstance(output, tp.Tuple):
                    args = output
                    kwargs = {}
                elif isinstance(output, tp.Dict):
                    args = ()
                    kwargs = output
                else:
                    args = (output,)
                    kwargs = {}
            if ctx is not None and utils.has_keyword_arg(f, "ctx"):
                kwargs["ctx"] = ctx

            output = f(*args, **kwargs)

        return output


class ModuleDefApply(tp.Protocol, tp.Generic[M]):
    def __call__(self, state: State, *states: State) -> ApplyCaller["PureModule[M]"]:
        ...


class TrainState(pytreelib.Pytree, tp.Generic[M]):
    def __init__(
        self,
        *,
        apply_fn: ModuleDefApply[M],
        params: State,
        tx: optax.GradientTransformation,
        step: int = 0,
        **kwargs,
    ):
        self.apply_fn = apply_fn
        self.params: State = containers.node(params)
        self.tx = tx
        self.opt_state = containers.node(tx.init(self.params))
        self.step = containers.node(step)
        for name, value in kwargs.items():
            setattr(self, name, value)

    if tp.TYPE_CHECKING:

        def __getattr__(self, key: str) -> tp.Any:
            ...

    def apply_gradients(self, grads: State, **kwargs) -> "TrainState[M]":
        updates, opt_state = self.tx.update(grads, self.opt_state, self.params)
        params = optax.apply_updates(self.params, updates)  # type: ignore
        step = self.step + 1
        return self.replace(
            params=params,
            opt_state=opt_state,
            step=step,
            **kwargs,
        )



import dataclasses
import typing as tp
from abc import ABCMeta
from typing import Any

import jax.tree_util as jtu

from nnx import errors, nodes, partitioning, reprlib, tracers
from nnx.containers import Container, Sharding, Variable
from nnx.state import State

A = tp.TypeVar("A")
M = tp.TypeVar("M", bound="Module")

Path = tp.Tuple[str, ...]
StateDict = tp.Dict[Path, tp.Any]
StateMapping = tp.Mapping[Path, tp.Any]


class ApplyCaller(tp.Protocol, tp.Generic[A]):
    def __getattr__(self, __name) -> "ApplyCaller[A]":
        ...

    def __getitem__(self, __name) -> "ApplyCaller[A]":
        ...

    def __call__(self, *args, **kwargs) -> tp.Tuple[tp.Any, A]:
        ...


@dataclasses.dataclass(repr=False)
class _SubmodulesRepr(reprlib.Representable):
    submodules: tp.Tuple[tp.Tuple[str, tp.Union["ModuleDef[Module]", int]], ...]

    def __nnx_repr__(self):
        yield reprlib.Object(type="", value_sep=", ")

        for name, submodule in self.submodules:
            yield reprlib.Attr(repr(name), submodule, start="(", end=")")


class ModuleDef(tp.Generic[M], reprlib.Representable):
    __slots__ = ("_type", "_index", "_submodules", "_static_fields")

    def __init__(
        self,
        type: tp.Type[M],
        index: int,
        submodules: tp.Tuple[tp.Tuple[str, tp.Union["ModuleDef[Module]", int]], ...],
        static_fields: tp.Tuple[tp.Tuple[str, tp.Any], ...],
    ):
        self._type = type
        self._index = index
        self._submodules = submodules
        self._static_fields = static_fields

    def __nnx_repr__(self):
        yield reprlib.Object(type=type(self))

        yield reprlib.Attr("type", self._type.__name__)
        yield reprlib.Attr("index", self._index)
        yield reprlib.Attr("submodules", _SubmodulesRepr(self._submodules))
        yield reprlib.Attr("static_fields", self._static_fields)

    def __hash__(self) -> int:
        return hash((self._type, self._submodules, self._static_fields))

    def __eq__(self, other: tp.Any) -> bool:
        if not isinstance(other, ModuleDef):
            return False
        return (
            self._type == other._type
            and self._submodules == other._submodules
            and self._static_fields == other._static_fields
        )

    @property
    def type(self) -> tp.Type[M]:
        return self._type

    @property
    def index(self) -> int:
        return self._index

    @property
    def submodules(
        self,
    ) -> tp.Tuple[tp.Tuple[str, tp.Union["ModuleDef[Module]", int]], ...]:
        return self._submodules

    @property
    def static_fields(self) -> tp.Tuple[tp.Tuple[str, tp.Any], ...]:
        return self._static_fields

    @tp.overload
    def merge(self, state: State, *states: State) -> M:
        ...

    def merge(self, *states: State) -> M:
        module = _build_module(self)
        current_state = State({})

        _update_module(module, current_state, states)

        return module

    def apply(self, state: State, *states: State) -> ApplyCaller["PureModule[M]"]:
        accessesor = DelayedAccessor()

        def _context(accessesor, *args, **kwargs) -> tp.Tuple[tp.Any, PureModule[M]]:
            module = self.merge(state, *states)
            fn = accessesor(module)
            out = fn(*args, **kwargs)
            return out, module.partition()

        return CallableProxy(_context, accessesor)  # type: ignore


def _moddef_flatten(moduledef: ModuleDef[M]):
    return (), (
        moduledef._type,
        moduledef._index,
        moduledef._submodules,
        moduledef._static_fields,
    )


def _moddef_unflatten(
    metadata: tp.Tuple[
        tp.Type[M],
        int,
        tp.Tuple[tp.Tuple[str, tp.Union["ModuleDef[Module]", int]], ...],
        tp.Tuple[tp.Tuple[str, tp.Any], ...],
    ],
    _,
) -> ModuleDef[M]:
    return ModuleDef(*metadata)


jtu.register_pytree_node(ModuleDef, _moddef_flatten, _moddef_unflatten)


class PureModule(tp.Tuple[State, ModuleDef[M]]):
    @classmethod
    def new(cls, state: State, moduledef: ModuleDef[M]) -> "PureModule[M]":
        return cls((state, moduledef))

    @property
    def state(self) -> State:
        return self[0]

    @property
    def moduledef(self) -> ModuleDef[M]:
        return self[1]

    def merge(self) -> M:
        return self.moduledef.merge(self.state)

    @property
    def apply(self) -> ApplyCaller["PureModule[M]"]:
        return self.moduledef.apply(self.state)

    @property
    def call(self) -> M:
        accessesor = DelayedAccessor()

        def _context(accessesor, *args, **kwargs):
            module = self.merge()
            fn = accessesor(module)
            return fn(*args, **kwargs)

        return CallableProxy(_context, accessesor)  # type: ignore

    def get_state(self) -> State:
        return self.state

    @tp.overload
    def filter(
        self,
        filter: partitioning.CollectionFilter,
        /,
    ) -> State:
        ...

    @tp.overload
    def filter(
        self,
        filter: partitioning.CollectionFilter,
        filter2: partitioning.CollectionFilter,
        /,
        *filters: partitioning.CollectionFilter,
    ) -> tp.Tuple[State, ...]:
        ...

    def filter(
        self, *filters: partitioning.CollectionFilter
    ) -> tp.Union[State, tp.Tuple[State, ...]]:
        return self.state.filter(*filters)

    @tp.overload
    def partition(self, first: partitioning.CollectionFilter, /) -> "PureModule[M]":
        ...

    @tp.overload
    def partition(
        self,
        first: partitioning.CollectionFilter,
        second: partitioning.CollectionFilter,
        /,
        *filters: partitioning.CollectionFilter,
    ) -> tp.Tuple[tp.Tuple[State, ...], ModuleDef[M]]:
        ...

    def partition(
        self,
        *filters: partitioning.CollectionFilter,
    ) -> tp.Union["PureModule[M]", tp.Tuple[tp.Tuple[State, ...], ModuleDef[M]],]:
        states = self.state.partition(*filters)
        if isinstance(states, State):
            return PureModule.new(states, self.moduledef)
        else:
            return states, self.moduledef

    @tp.overload
    def pop_state(
        self, filter: partitioning.CollectionFilter, /
    ) -> tp.Tuple[State, "PureModule[M]"]:
        ...

    @tp.overload
    def pop_state(
        self,
        filter: partitioning.CollectionFilter,
        filter2: partitioning.CollectionFilter,
        /,
        *filters: partitioning.CollectionFilter,
    ) -> tp.Tuple[tp.Tuple[State, ...], "PureModule[M]"]:
        ...

    def pop_state(
        self, *filters: partitioning.CollectionFilter
    ) -> tp.Tuple[tp.Union[State, tp.Tuple[State, ...]], "PureModule[M]"]:
        if len(filters) == 0:
            raise ValueError("At least one filter must be provided")
        else:
            *states, rest = self.state.partition(*filters, ...)

        if len(states) == 1:
            states = states[0]
        else:
            states = tuple(states)

        return states, PureModule.new(rest, self.moduledef)

    def update_state(
        self,
        updates: tp.Union[M, "PureModule[M]", State, tp.Tuple[State, ...]],
    ) -> "PureModule[M]":
        if isinstance(updates, Module):
            states = (updates.partition()[0],)
        elif isinstance(updates, PureModule):
            states = (updates.state,)
        elif isinstance(updates, State):
            states = (updates,)
        elif isinstance(updates, tuple):
            states = updates
        else:
            raise TypeError(
                f"Expected Module, PureModule, State or tuple of State, "
                f"got {type(updates).__name__}"
            )

        state = State.merge(*states)
        return PureModule.new(state, self.moduledef)


def _pure_module_flatten(bounded: PureModule[M]):
    return tuple(bounded), None


def _pure_module_unflatten(_, values: tp.Tuple[State, ModuleDef[M]]):
    return PureModule(values)


jtu.register_pytree_node(PureModule, _pure_module_flatten, _pure_module_unflatten)


class _ProxyContext(tp.Protocol):
    def __call__(self, __fn: tp.Callable[..., tp.Any], *args, **kwargs) -> tp.Any:
        ...


@dataclasses.dataclass
class CallableProxy:
    _proxy_context: _ProxyContext
    _proxy_callable: tp.Callable[..., tp.Any]

    def __call__(self, *args, **kwargs):
        return self._proxy_context(self._proxy_callable, *args, **kwargs)

    def __getattr__(self, name) -> "CallableProxy":
        return CallableProxy(self._proxy_context, getattr(self._proxy_callable, name))

    def __getitem__(self, key) -> "CallableProxy":
        return CallableProxy(self._proxy_context, self._proxy_callable[key])


def _identity(x):
    return x


@dataclasses.dataclass
class DelayedAccessor:
    accessor: tp.Callable[[tp.Any], tp.Any] = _identity

    def __call__(self, x):
        return self.accessor(x)

    def __getattr__(self, name):
        return DelayedAccessor(lambda x: getattr(x, name))

    def __getitem__(self, key):
        return DelayedAccessor(lambda x: x[key])


SEEN_MODULES_REPR: tp.Set[int] = set()


class ModuleState(reprlib.Representable):
    __slots__ = ("_trace_state",)

    def __init__(self, trace_state: tracers.TraceState):
        self._trace_state = trace_state

    @property
    def trace_state(self) -> tracers.TraceState:
        return self._trace_state

    def __nnx_repr__(self):
        yield reprlib.Object(type(self))
        yield reprlib.Attr("trace_state", self._trace_state)


class ModuleMeta(ABCMeta):
    if not tp.TYPE_CHECKING:

        def __call__(self, *args: Any, **kwargs: Any) -> Any:
            return self.call(*args, **kwargs)

    def call(self: tp.Type[M], *args, **kwargs) -> M:
        module = self.__new__(self, *args, **kwargs)
        vars(module)["_module__state"] = ModuleState(tracers.TraceState())
        module.__init__(*args, **kwargs)

        if dataclasses.is_dataclass(module):
            assert isinstance(module, Module)
            for field in dataclasses.fields(module):
                if "nnx_container_fn" not in field.metadata:
                    continue

                container_fn = field.metadata["nnx_container_fn"]
                value = vars(module)[field.name]
                value = container_fn(value)
                vars(module)[field.name] = value

        return module


class Module(reprlib.Representable, metaclass=ModuleMeta):
    if tp.TYPE_CHECKING:
        _module__state: ModuleState

    if not tp.TYPE_CHECKING:

        def __getattribute__(self, name: str) -> Any:
            value = object.__getattribute__(self, name)
            if isinstance(value, Container):
                return value.value
            return value

        def __setattr__(self, name: str, value: Any) -> None:
            self._setattr(name, value)

    def _setattr(self, name: str, value: Any) -> None:
        if not self._module__state.trace_state.is_valid():
            raise errors.TraceContextError(
                "Cannot mutate Module from different trace level"
            )

        vars_dict = vars(self)
        if name in vars_dict and isinstance(vars_dict[name], Container):
            vars_dict[name] = vars_dict[name].replace_value(value)
        else:
            if isinstance(value, Container):
                value = value.copy()
            vars_dict[name] = value

    def __hash__(self) -> int:
        return id(self)

    def __nnx_repr__(self):
        if id(self) in SEEN_MODULES_REPR:
            yield reprlib.Object(type=type(self), empty_repr="...")
            return

        yield reprlib.Object(type=type(self))
        SEEN_MODULES_REPR.add(id(self))

        try:
            for name, value in vars(self).items():
                if isinstance(value, Module) or (
                    not nodes.is_node(value) and not name.startswith("_")
                ):
                    yield reprlib.Attr(name, value)
        finally:
            SEEN_MODULES_REPR.remove(id(self))

    def clone(self: M) -> M:
        return self.partition().merge()

    @tp.overload
    def partition(self: M) -> PureModule[M]:
        ...

    @tp.overload
    def partition(self: M, first: partitioning.CollectionFilter, /) -> PureModule[M]:
        ...

    @tp.overload
    def partition(
        self: M,
        first: partitioning.CollectionFilter,
        second: partitioning.CollectionFilter,
        /,
        *filters: partitioning.CollectionFilter,
    ) -> tp.Tuple[tp.Tuple[State, ...], ModuleDef[M]]:
        ...

    def partition(
        self: M,
        *filters: partitioning.CollectionFilter,
    ) -> tp.Union[PureModule[M], tp.Tuple[tp.Tuple[State, ...], ModuleDef[M]]]:
        moduledef = _get_module_def(self)
        state = _get_module_state(self)

        if len(filters) == 0:
            states = state
        elif len(filters) == 1:
            states = state.partition(filters[0])
        else:
            states = state.partition(filters[0], filters[1], *filters[2:])

        if isinstance(states, tuple):
            return states, moduledef
        else:
            return PureModule.new(states, moduledef)

    def get_state(self) -> State:
        return _get_module_state(self)

    @tp.overload
    def filter(self, first: partitioning.CollectionFilter, /) -> State:
        ...

    @tp.overload
    def filter(
        self,
        first: partitioning.CollectionFilter,
        second: partitioning.CollectionFilter,
        /,
        *filters: partitioning.CollectionFilter,
    ) -> tp.Tuple[State, ...]:
        ...

    def filter(
        self,
        first: partitioning.CollectionFilter,
        /,
        *filters: partitioning.CollectionFilter,
    ) -> tp.Union[State, tp.Tuple[State, ...]]:
        state = _get_module_state(self)

        if len(filters) == 0:
            states = state.filter(first)
        else:
            states = state.filter(first, filters[0], *filters[1:])

        return states

    @tp.overload
    def pop_state(
        self,
        filter: partitioning.CollectionFilter,
        /,
    ) -> State:
        ...

    @tp.overload
    def pop_state(
        self,
        filter: partitioning.CollectionFilter,
        filter2: partitioning.CollectionFilter,
        /,
        *filters: partitioning.CollectionFilter,
    ) -> tp.Tuple[State, ...]:
        ...

    def pop_state(
        self, *filters: partitioning.CollectionFilter
    ) -> tp.Union[State, tp.Tuple[State, ...]]:
        if len(filters) == 0:
            raise ValueError("Expected at least one filter")

        states = _pop(self, filters)

        if len(states) == 1:
            return states[0]
        else:
            return states

    @property
    def apply(self: M) -> ApplyCaller[M]:
        accessesor = DelayedAccessor()

        def _context(accessesor, *args, **kwargs) -> tp.Tuple[tp.Any, M]:
            module = self.clone()
            fn = accessesor(module)
            out = fn(*args, **kwargs)
            return out, module

        return CallableProxy(_context, accessesor)  # type: ignore

    def update_state(
        self: M,
        updates: tp.Union[M, PureModule[M], State, tp.Tuple[State, ...]],
    ) -> None:
        current_state = _get_module_state(self)

        if isinstance(updates, PureModule):
            updates = updates.state
        elif isinstance(updates, Module):
            assert type(self) == type(updates)
            updates = _get_module_state(updates)

        _update_module(self, current_state, updates)

    @tp.overload
    def mutable_state_dict(self) -> tp.Dict[Path, "MutableLeaf"]:
        ...

    @tp.overload
    def mutable_state_dict(self, sep: str) -> tp.Dict[str, "MutableLeaf"]:
        ...

    def mutable_state_dict(
        self, sep: tp.Optional[str] = None
    ) -> tp.Union[tp.Dict[Path, "MutableLeaf"], tp.Dict[str, "MutableLeaf"]]:
        mutable_leaves = (
            (path, MutableLeaf(self, path)) for path, _ in _iter_state(self)
        )
        if sep is not None:
            state = {sep.join(path): leaf for path, leaf in mutable_leaves}
        else:
            state = {path: leaf for path, leaf in mutable_leaves}

        return state

    # Pytree Definition
    # def __init_subclass__(cls) -> None:
    #     super().__init_subclass__()

    #     def _flatten(module: Module, *, with_keys: bool):
    #         state, moduledef = module.partition()
    #         paths = tuple(state.keys())

    #         if with_keys:
    #             nodes = tuple(
    #                 (jtu.DictKey(path), value) for path, value in state.items()
    #             )
    #         else:
    #             nodes = tuple(state.values())

    #         return nodes, (paths, moduledef)

    #     def _unflatten(
    #         paths_moddef: tp.Tuple[tp.Tuple[Path, ...], ModuleDef[M]],
    #         nodes: tp.Tuple[tp.Any, ...],
    #     ) -> M:
    #         paths, moduledef = paths_moddef
    #         return moduledef.merge(State(zip(paths, nodes)))

    #     jtu.register_pytree_with_keys(
    #         cls,
    #         partial(_flatten, with_keys=True),
    #         _unflatten,
    #         flatten_func=partial(_flatten, with_keys=False),
    #     )


class MutableLeaf(reprlib.Representable):
    __slots__ = ("_module", "_name")

    def __init__(self, root: Module, path: Path):
        *module_path, name = path
        module = _get_value_path(root, module_path)
        if not isinstance(module, Module):
            raise ValueError(
                f"Expected a module at path {path[:-1]}, ",
                f" got {type(module).__name__}",
            )
        self._module = module
        self._name = name

    @property
    def value(self) -> tp.Any:
        return getattr(self._module, self._name)

    @value.setter
    def value(self, value: tp.Any) -> None:
        setattr(self._module, self._name, value)

    @property
    def collection(self) -> tp.Optional[str]:
        module = vars(self._module)[self._name]
        if not isinstance(module, Variable):
            return None

        return module.collection

    @property
    def sharding(self) -> tp.Optional[Sharding]:
        module = vars(self._module)[self._name]
        if not isinstance(module, Variable):
            return None

        return module.sharding


def _get_module_state(module: Module) -> State:
    return State(_iter_state(module))


def _get_module_def(module: M) -> ModuleDef[M]:
    module_index: tp.Dict[int, int] = {}
    path: Path = ()

    moduledef = _make_module_def_recursive(module, module_index, path)
    assert isinstance(moduledef, ModuleDef)

    return moduledef


def _make_module_def_recursive(
    module: M,
    module_index: tp.Dict[int, int],
    path: Path,
) -> tp.Union[ModuleDef[M], int]:
    if id(module) in module_index:
        return module_index[id(module)]

    index = len(module_index)
    module_index[id(module)] = index

    submodules = []
    static_fields = []

    for name, value in sorted(vars(module).items(), key=lambda x: x[0]):
        value_path = (*path, name)
        if isinstance(value, Module):
            submodule_def = _make_module_def_recursive(value, module_index, value_path)
            submodules.append((name, submodule_def))
        elif not nodes.is_node(value) and not name.startswith("_module__"):
            static_fields.append((name, value))

    module_def = ModuleDef(
        type=type(module),
        index=index,
        submodules=tuple(submodules),
        static_fields=tuple(static_fields),
    )
    return module_def


def _iter_state(module: Module) -> tp.Iterator[tp.Tuple[Path, tp.Any]]:
    seen_modules: tp.Set[int] = set()
    path: Path = ()

    yield from _iter_state_recursive(module, seen_modules, path)


def _iter_state_recursive(
    module: Module, seen_modules: tp.Set[int], path: Path
) -> tp.Iterator[tp.Tuple[Path, tp.Any]]:
    if id(module) in seen_modules:
        return

    seen_modules.add(id(module))

    for name, value in sorted(vars(module).items(), key=lambda x: x[0]):
        value_path = (*path, name)
        if isinstance(value, Module):
            yield from _iter_state_recursive(value, seen_modules, value_path)
        elif nodes.is_node(value):
            yield value_path, value


def _set_value_at_path(module: tp.Any, path: tp.Sequence[str], value: tp.Any):
    if len(path) == 1:
        setattr(module, path[0], value)
    else:
        _set_value_at_path(vars(module)[path[0]], path[1:], value)


def _get_value_path(module: tp.Any, path: tp.Sequence[str]) -> tp.Any:
    if len(path) == 0:
        return module
    else:
        return _get_value_path(vars(module)[path[0]], path[1:])


def _build_module(moduledef: ModuleDef[M]) -> M:
    index_module: tp.Dict[int, Module] = {}
    module = _build_module_recursive(moduledef, index_module)
    return module


def _build_module_recursive(
    moduledef: tp.Union[ModuleDef[M], int],
    index_module: tp.Dict[int, Module],
) -> M:
    if isinstance(moduledef, int):
        return index_module[moduledef]  # type: ignore

    assert moduledef.index not in index_module

    # add a dummy module to the index to avoid infinite recursion
    module = object.__new__(moduledef.type)
    index_module[moduledef.index] = module

    submodules = {
        name: _build_module_recursive(submodule, index_module)
        for name, submodule in moduledef.submodules
    }

    vars(module).update(moduledef.static_fields)
    vars(module).update(submodules)
    vars(module)["_module__state"] = ModuleState(tracers.TraceState())

    return module


def _pop(
    module: Module,
    filters: tp.Tuple[partitioning.CollectionFilter, ...],
) -> tp.Tuple[State, ...]:
    module_index: tp.Dict[int, int] = {}
    path: Path = ()
    predicates = tuple(partitioning.to_predicate(filter) for filter in filters)
    states = tuple({} for _ in predicates)
    _pop_recursive(module, module_index, path, states, predicates)

    return tuple(State(x) for x in states)


def _pop_recursive(
    module: Module,
    module_index: tp.Dict[int, int],
    path: Path,
    states: tp.Tuple[tp.Dict[Path, tp.Any]],
    predicates: tp.Tuple[partitioning.Predicate, ...],
) -> None:
    if id(module) in module_index:
        return

    for name, value in list(vars(module).items()):
        value_path = (*path, name)
        if isinstance(value, Module):
            _pop_recursive(value, module_index, value_path, states, predicates)
            continue
        elif not nodes.is_node(value):
            continue

        for state, predicate in zip(states, predicates):
            if predicate(value_path, value):
                state[value_path] = value
                delattr(module, name)
                break

    module_index[id(module)] = len(module_index)


def _update_module(
    module: Module, current_state: State, updates: tp.Union[State, tp.Tuple[State, ...]]
) -> None:
    new_states = [current_state]

    if isinstance(updates, State):
        new_states.append(updates)
    else:
        new_states.extend(updates)

    state: StateDict = {}
    for new_state in new_states:
        state.update(new_state)

    for path, value in state.items():
        _set_value_at_path(module, path, value)


# register nodes
nodes.register_node_type(Module)
nodes.register_node_type(PureModule)



import typing as tp

import jax
import jax.numpy as jnp
from jax import lax

import nnx
from nnx import contextlib, utils
from nnx.module import Module
from nnx.nn import dtypes, initializers

PRNGKey = jax.Array
Array = jax.Array
Shape = tp.Tuple[int, ...]
Dtype = tp.Any  # this could be a real type?

Axes = tp.Union[int, tp.Any]


def _canonicalize_axes(rank: int, axes: Axes) -> tp.Tuple[int, ...]:
    """Returns a tuple of deduplicated, sorted, and positive axes."""
    if not isinstance(axes, tp.Iterable):
        axes = (axes,)
    return tuple(set([rank + axis if axis < 0 else axis for axis in axes]))


def _abs_sq(x):
    """Computes the elementwise square of the absolute value |x|^2."""
    if jnp.iscomplexobj(x):
        return lax.square(lax.real(x)) + lax.square(lax.imag(x))
    else:
        return lax.square(x)


def _compute_stats(
    x: Array,
    axes: tp.Optional[Axes],
    dtype: tp.Optional[Dtype],
    axis_name: tp.Optional[str] = None,
    axis_index_groups: tp.Any = None,
    use_mean: bool = True,
):
    """Computes mean and variance statistics.

    This implementation takes care of a few important details:
    - Computes in float32 precision for stability in half precision training.
    - mean and variance are computable in a single XLA fusion,
      by using Var = E[|x|^2] - |E[x]|^2 instead of Var = E[|x - E[x]|^2]).
    - Clips negative variances to zero which can happen due to
      roundoff errors. This avoids downstream NaNs.
    - Supports averaging across a parallel axis and subgroups of a parallel axis
      with a single `lax.pmean` call to avoid latency.

    Arguments:
      x: Input array.
      axes: The axes in ``x`` to compute mean and variance statistics for.
      dtype: tp.Optional dtype specifying the minimal precision. Statistics
        are always at least float32 for stability (default: dtype of x).
      axis_name: tp.Optional name for the pmapped axis to compute mean over.
      axis_index_groups: tp.Optional axis indices.
      use_mean: If true, calculate the mean from the input and use it when
        computing the variance. If false, set the mean to zero and compute
        the variance without subtracting the mean.

    Returns:
      A pair ``(mean, var)``.
    """
    if dtype is None:
        dtype = jnp.result_type(x)
    # promote x to at least float32, this avoids half precision computation
    # but preserves double or complex floating points
    dtype = jnp.promote_types(dtype, jnp.float32)
    x = jnp.asarray(x, dtype)

    mean2 = jnp.mean(_abs_sq(x), axes)
    if use_mean:
        mean = jnp.mean(x, axes)
    else:
        mean = jnp.zeros(mean2.shape, dtype=dtype)

    if axis_name is not None:
        concatenated_mean = jnp.concatenate([mean, mean2])
        mean, mean2 = jnp.split(
            lax.pmean(
                concatenated_mean,
                axis_name=axis_name,
                axis_index_groups=axis_index_groups,
            ),
            2,
        )
    # mean2 - _abs_sq(mean) is not guaranteed to be non-negative due
    # to floating point round-off errors.
    var = jnp.maximum(0.0, mean2 - _abs_sq(mean))
    return mean, var


def _normalize(
    x: Array,
    mean: Array,
    var: Array,
    scale: tp.Optional[Array],
    bias: tp.Optional[Array],
    reduction_axes: Axes,
    feature_axes: Axes,
    dtype: Dtype,
    epsilon: float,
):
    """ "Normalizes the input of a normalization layer and optionally applies a learned scale and bias.

    Arguments:
      x: The input.
      mean: Mean to use for normalization.
      var: Variance to use for normalization.
      reduction_axes: The axes in ``x`` to reduce.
      feature_axes: Axes containing features. A separate bias and scale is learned
        for each specified feature.
      dtype: The dtype of the result (default: infer from input and params).
      epsilon: Normalization epsilon.

    Returns:
      The normalized input.
    """
    reduction_axes = _canonicalize_axes(x.ndim, reduction_axes)
    feature_axes = _canonicalize_axes(x.ndim, feature_axes)
    stats_shape = list(x.shape)
    for axis in reduction_axes:
        stats_shape[axis] = 1
    mean = mean.reshape(stats_shape)
    var = var.reshape(stats_shape)
    feature_shape = [1] * x.ndim
    reduced_feature_shape = []
    for ax in feature_axes:
        feature_shape[ax] = x.shape[ax]
        reduced_feature_shape.append(x.shape[ax])
    y = x - mean
    mul = lax.rsqrt(var + epsilon)
    args = [x]
    if scale is not None:
        scale = scale.reshape(feature_shape)
        mul *= scale
        args.append(scale)
    y *= mul
    if bias is not None:
        bias = bias.reshape(feature_shape)
        y += bias
        args.append(bias)
    dtype = dtypes.canonicalize_dtype(*args, dtype=dtype)
    return jnp.asarray(y, dtype)


class BatchNorm(Module):
    """BatchNorm Module.

    Attributes:
      use_running_average: if True, the statistics stored in batch_stats
        will be used instead of computing the batch statistics on the input.
      axis: the feature or non-batch axis of the input.
      momentum: decay rate for the exponential moving average of
        the batch statistics.
      epsilon: a small float added to variance to avoid dividing by zero.
      dtype: the dtype of the result (default: infer from input and params).
      param_dtype: the dtype passed to parameter initializers (default: float32).
      use_bias:  if True, bias (beta) is added.
      use_scale: if True, multiply by scale (gamma).
        When the next layer is linear (also e.g. nn.relu), this can be disabled
        since the scaling will be done by the next layer.
      bias_init: initializer for bias, by default, zero.
      scale_init: initializer for scale, by default, one.
      axis_name: the axis name used to combine batch statistics from multiple
        devices. See `jax.pmap` for a description of axis names (default: None).
      axis_index_groups: groups of axis indices within that named axis
        representing subsets of devices to reduce over (default: None). For
        example, `[[0, 1], [2, 3]]` would independently batch-normalize over
        the examples on the first two and last two devices. See `jax.lax.psum`
        for more details.
    """

    def __init__(
        self,
        num_features: int,
        *,
        use_running_average: tp.Optional[bool] = None,
        axis: int = -1,
        momentum: float = 0.99,
        epsilon: float = 1e-5,
        dtype: tp.Optional[Dtype] = None,
        param_dtype: Dtype = jnp.float32,
        use_bias: bool = True,
        use_scale: bool = True,
        bias_init: tp.Callable[[PRNGKey, Shape, Dtype], Array] = initializers.zeros(),
        scale_init: tp.Callable[[PRNGKey, Shape, Dtype], Array] = initializers.ones(),
        axis_name: tp.Optional[str] = None,
        axis_index_groups: tp.Any = None,
        ctx: contextlib.Context,
    ):
        feature_shape = (num_features,)
        self.mean = nnx.var("batch_stats", jnp.zeros(feature_shape, jnp.float32))
        self.var = nnx.var("batch_stats", jnp.ones(feature_shape, jnp.float32))

        if use_scale:
            key = ctx.make_rng("params")
            self.scale = nnx.param(scale_init(key, feature_shape, param_dtype))
        else:
            self.scale = nnx.param(None)

        if use_bias:
            key = ctx.make_rng("params")
            self.bias = nnx.param(bias_init(key, feature_shape, param_dtype))
        else:
            self.bias = nnx.param(None)

        self.num_features = num_features
        self.use_running_average = use_running_average
        self.axis = axis
        self.momentum = momentum
        self.epsilon = epsilon
        self.dtype = dtype
        self.param_dtype = param_dtype
        self.use_bias = use_bias
        self.use_scale = use_scale
        self.bias_init = bias_init
        self.scale_init = scale_init
        self.axis_name = axis_name
        self.axis_index_groups = axis_index_groups

    def __call__(
        self,
        x,
        use_running_average: tp.Optional[bool] = None,
        *,
        ctx: tp.Optional[contextlib.Context] = None,
    ):
        """Normalizes the input using batch statistics.

        Args:
          x: the input to be normalized.
          use_running_average: if true, the statistics stored in batch_stats
            will be used instead of computing the batch statistics on the input.

        Returns:
          Normalized inputs (the same shape as inputs).
        """

        use_running_average = utils.first_from(
            use_running_average,
            self.use_running_average,
            ctx and ctx.get_flag("use_running_average"),
        )
        feature_axes = _canonicalize_axes(x.ndim, self.axis)
        reduction_axes = tuple(i for i in range(x.ndim) if i not in feature_axes)

        if use_running_average:
            mean, var = self.mean, self.var
        else:
            mean, var = _compute_stats(
                x,
                reduction_axes,
                dtype=self.dtype,
                axis_name=self.axis_name,
                axis_index_groups=self.axis_index_groups,
            )

            self.mean = self.momentum * self.mean + (1 - self.momentum) * mean
            self.var = self.momentum * self.var + (1 - self.momentum) * var

        return _normalize(
            x,
            mean,
            var,
            self.scale,
            self.bias,
            reduction_axes,
            feature_axes,
            self.dtype,
            self.epsilon,
        )


class LayerNorm(Module):
    """Layer normalization (https://arxiv.org/abs/1607.06450).

    LayerNorm normalizes the activations of the layer for each given example in a
    batch independently, rather than across a batch like Batch Normalization.
    i.e. applies a transformation that maintains the mean activation within
    each example close to 0 and the activation standard deviation close to 1.

    Attributes:
        epsilon: A small float added to variance to avoid dividing by zero.
        dtype: the dtype of the result (default: infer from input and params).
        param_dtype: the dtype passed to parameter initializers (default: float32).
        use_bias:  If True, bias (beta) is added.
        use_scale: If True, multiply by scale (gamma). When the next layer is linear
            (also e.g. nn.relu), this can be disabled since the scaling will be done
            by the next layer.
        bias_init: Initializer for bias, by default, zero.
        scale_init: Initializer for scale, by default, one.
        reduction_axes: Axes for computing normalization statistics.
        feature_axes: Feature axes for learned bias and scaling.
        axis_name: the axis name used to combine batch statistics from multiple
            devices. See `jax.pmap` for a description of axis names (default: None).
            This is only needed if the model is subdivided across devices, i.e. the
            array being normalized is sharded across devices within a pmap.
        axis_index_groups: groups of axis indices within that named axis
            representing subsets of devices to reduce over (default: None). For
            example, `[[0, 1], [2, 3]]` would independently batch-normalize over
            the examples on the first two and last two devices. See `jax.lax.psum`
            for more details.
    """

    def __init__(
        self,
        num_features: int,
        *,
        epsilon: float = 1e-6,
        dtype: tp.Optional[Dtype] = None,
        param_dtype: Dtype = jnp.float32,
        use_bias: bool = True,
        use_scale: bool = True,
        bias_init: tp.Callable[[PRNGKey, Shape, Dtype], Array] = initializers.zeros(),
        scale_init: tp.Callable[[PRNGKey, Shape, Dtype], Array] = initializers.ones(),
        reduction_axes: Axes = -1,
        feature_axes: Axes = -1,
        axis_name: tp.Optional[str] = None,
        axis_index_groups: tp.Any = None,
        ctx: contextlib.Context,
    ):
        feature_shape = (num_features,)

        if use_scale:
            key = ctx.make_rng("params")
            self.scale = nnx.param(scale_init(key, feature_shape, param_dtype))
        else:
            self.scale = nnx.param(None)

        if use_bias:
            key = ctx.make_rng("params")
            self.bias = nnx.param(bias_init(key, feature_shape, param_dtype))
        else:
            self.bias = nnx.param(None)

        self.num_features = num_features
        self.epsilon = epsilon
        self.dtype = dtype
        self.param_dtype = param_dtype
        self.use_bias = use_bias
        self.use_scale = use_scale
        self.bias_init = bias_init
        self.scale_init = scale_init
        self.reduction_axes = reduction_axes
        self.feature_axes = feature_axes
        self.axis_name = axis_name
        self.axis_index_groups = axis_index_groups

    def __call__(self, x):
        """Applies layer normalization on the input.

        Args:
          x: the inputs

        Returns:
          Normalized inputs (the same shape as inputs).
        """
        mean, var = _compute_stats(
            x, self.reduction_axes, self.dtype, self.axis_name, self.axis_index_groups
        )

        return _normalize(
            x,
            mean,
            var,
            self.scale,
            self.bias,
            self.reduction_axes,
            self.feature_axes,
            self.dtype,
            self.epsilon,
        )



from jax.nn import (
    celu,
    elu,
    gelu,
    glu,
    hard_sigmoid,
    hard_silu,
    hard_swish,
    hard_tanh,
    leaky_relu,
    log_sigmoid,
    log_softmax,
    logsumexp,
    normalize,
    one_hot,
    relu,
    relu6,
    selu,
    sigmoid,
    silu,
    soft_sign,
    softmax,
    softplus,
    standardize,
    swish,
)
from jax.numpy import tanh

__all__ = [
    "celu",
    "elu",
    "gelu",
    "glu",
    "hard_sigmoid",
    "hard_silu",
    "hard_swish",
    "hard_tanh",
    "leaky_relu",
    "log_sigmoid",
    "log_softmax",
    "logsumexp",
    "normalize",
    "one_hot",
    "relu",
    "relu6",
    "selu",
    "sigmoid",
    "silu",
    "soft_sign",
    "softmax",
    "softplus",
    "standardize",
    "swish",
    "tanh",
]



import dataclasses
import typing as tp

import typing_extensions as tpe

from nnx import containers

A = tp.TypeVar("A")


def field(
    *,
    default: tp.Any = dataclasses.MISSING,
    default_factory: tp.Any = dataclasses.MISSING,
    init: bool = True,
    repr: bool = True,
    hash: tp.Optional[bool] = None,
    compare: bool = True,
    metadata: tp.Optional[tp.Mapping[str, tp.Any]] = None,
):
    return dataclasses.field(  # type: ignore
        default=default,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )


def node_field(
    *,
    default: tp.Any = dataclasses.MISSING,
    default_factory: tp.Any = dataclasses.MISSING,
    init: bool = True,
    repr: bool = True,
    hash: tp.Optional[bool] = None,
    compare: bool = True,
    metadata: tp.Optional[tp.Mapping[str, tp.Any]] = None,
):
    if metadata is None:
        metadata = {}
    else:
        metadata = dict(metadata)

    if "nnx_container_fn" in metadata:
        raise ValueError("'nnx_container_fn' found in metadata")

    metadata["nnx_container_fn"] = lambda value: containers.node(value)

    return field(
        default=default,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )


def static_field(
    *,
    default: tp.Any = dataclasses.MISSING,
    default_factory: tp.Any = dataclasses.MISSING,
    init: bool = True,
    repr: bool = True,
    hash: tp.Optional[bool] = None,
    compare: bool = True,
    metadata: tp.Optional[tp.Mapping[str, tp.Any]] = None,
):
    if metadata is None:
        metadata = {}
    else:
        metadata = dict(metadata)

    if "nnx_container_fn" in metadata:
        raise ValueError("'nnx_container_fn' found in metadata")

    metadata["nnx_container_fn"] = lambda value: containers.static(value)

    return field(
        default=default,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )


def var_field(
    collection: str,
    *,
    default: tp.Any = dataclasses.MISSING,
    default_factory: tp.Any = dataclasses.MISSING,
    init: bool = True,
    repr: bool = True,
    hash: tp.Optional[bool] = None,
    compare: bool = True,
    metadata: tp.Optional[tp.Mapping[str, tp.Any]] = None,
    sharding: tp.Optional[containers.Sharding] = None,
) -> tp.Any:
    if metadata is None:
        metadata = {}
    else:
        metadata = dict(metadata)

    if "nnx_container_fn" in metadata:
        raise ValueError("'nnx_container_fn' found in metadata")

    metadata["nnx_container_fn"] = lambda value: containers.var(
        collection, value, sharding=sharding
    )

    return field(
        default=default,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )


def param_field(
    default: tp.Any = dataclasses.MISSING,
    *,
    default_factory: tp.Any = dataclasses.MISSING,
    init: bool = True,
    repr: bool = True,
    hash: tp.Optional[bool] = None,
    compare: bool = True,
    metadata: tp.Optional[tp.Mapping[str, tp.Any]] = None,
) -> tp.Any:
    return var_field(
        "params",
        default=default,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )


@tp.overload
def dataclass(cls: tp.Type[A]) -> tp.Type[A]:
    ...


@tp.overload
def dataclass(
    *,
    init: bool = True,
    repr: bool = True,
    eq: bool = True,
    order: bool = False,
    unsafe_hash: bool = False,
    frozen: bool = False,
) -> tp.Callable[[tp.Type[A]], tp.Type[A]]:
    ...


@tpe.dataclass_transform(
    field_specifiers=(field, node_field, static_field, var_field, param_field)
)
def dataclass(
    cls: tp.Optional[tp.Type[A]] = None,
    init: bool = True,
    repr: bool = True,
    eq: bool = True,
    order: bool = False,
    unsafe_hash: bool = False,
    frozen: bool = False,
) -> tp.Union[tp.Type[A], tp.Callable[[tp.Type[A]], tp.Type[A]]]:
    decorator = dataclasses.dataclass(
        init=init,
        repr=repr,
        eq=eq,
        order=order,
        unsafe_hash=unsafe_hash,
        frozen=frozen,
    )

    if cls is None:
        return decorator

    return decorator(cls)



import inspect
import typing as tp

A = tp.TypeVar("A")


def first_from(*args: tp.Optional[A]) -> A:
    """Return the first non-None argument."""
    for arg in args:
        if arg is not None:
            return arg
    raise ValueError("No non-None arguments found.")


def has_keyword_arg(func: tp.Callable[..., tp.Any], name: str) -> bool:
    """Return True if func has keyword-only arguments with the given name."""
    return any(
        param.name == name
        and param.kind in (param.KEYWORD_ONLY, param.POSITIONAL_OR_KEYWORD)
        for param in inspect.signature(func).parameters.values()
    )



import builtins
import dataclasses
import typing as tp

import jax
import numpy as np

import nnx

if tp.TYPE_CHECKING:
    ellipsis = builtins.ellipsis
else:
    ellipsis = tp.Any

FilterLiteral = tp.Union[str, type]
Path = tp.Tuple[str, ...]
Predicate = tp.Callable[[Path, tp.Any], bool]
CollectionFilter = tp.Union[
    FilterLiteral, tp.Sequence[FilterLiteral], Predicate, ellipsis
]


def to_predicate(collection_filter: CollectionFilter) -> Predicate:
    if isinstance(collection_filter, str):
        return FromCollection(collection_filter)
    elif isinstance(collection_filter, type):
        return OfType(collection_filter)
    elif collection_filter is Ellipsis:
        return Everything()
    elif callable(collection_filter):
        return collection_filter
    elif isinstance(collection_filter, tp.Sequence):
        return Any(collection_filter)
    else:
        raise TypeError(f"Invalid collection filter: {collection_filter}")


@dataclasses.dataclass
class FromCollection:
    collection: str

    def __call__(self, path: Path, x: tp.Any):
        if isinstance(x, nnx.Variable):
            return x.collection == self.collection
        return False


@dataclasses.dataclass
class OfType:
    type: type

    def __call__(self, path: Path, x: tp.Any):
        return isinstance(x, self.type)


class Any:
    def __init__(self, collection_filters: tp.Sequence[CollectionFilter]):
        self.predicates = tuple(
            to_predicate(collection_filter) for collection_filter in collection_filters
        )

    def __call__(self, path: Path, x: tp.Any):
        return any(predicate(path, x) for predicate in self.predicates)


class Not:
    def __init__(self, collection_filter: CollectionFilter):
        self.predicate = to_predicate(collection_filter)

    def __call__(self, path: Path, x: tp.Any):
        return not self.predicate(path, x)


class Everything:
    def __call__(self, path: Path, x: tp.Any):
        return True


class NonVariable:
    def __call__(self, path: Path, x: tp.Any):
        return not isinstance(x, nnx.Variable)


class Buffers:
    def __call__(self, path: Path, x: tp.Any):
        return isinstance(x, (jax.Array, np.ndarray))


buffers = Buffers()



import typing as tp

import jax
import numpy as np

node_types: tp.Tuple[type, ...] = ()


def register_node_type(node_type: type) -> None:
    global node_types
    node_types += (node_type,)


def is_node(obj: object) -> bool:
    return isinstance(obj, node_types)


# register nodes
register_node_type(jax.Array)
register_node_type(np.ndarray)



# Taken from flax/core/tracer.py 🏴‍☠️

import contextlib
import dataclasses
import threading
import typing as tp

import jax
import jax.core
from jax.core import MainTrace

from nnx import reprlib


@tp.runtime_checkable
class Tracer(tp.Protocol):
    _trace: jax.core.Trace


def get_top_trace(pytree: tp.Union[tp.Any, Tracer]) -> MainTrace:
    """Returns the main top trace of a sequence of tracers."""
    if isinstance(pytree, Tracer):
        return pytree._trace.main

    return jax.core.find_top_trace(jax.tree_util.tree_leaves(pytree)).main


def current_jax_trace() -> MainTrace:
    """Returns the innermost Jax tracer."""
    return get_top_trace(())


def get_all_traces(pytree: tp.Union[tp.Any, Tracer]) -> tp.Set[MainTrace]:
    """Returns True if all tracers have the same main trace."""
    if isinstance(pytree, Tracer):
        return {pytree._trace.main}
    else:
        return {
            trace._trace.main
            for trace in jax.tree_util.tree_leaves(pytree)
            if isinstance(trace, Tracer)
        }


def trace_level(main):
    """Returns the level of the trace of -infinity if it is None."""
    if main:
        return main.level
    return float("-inf")


@dataclasses.dataclass
class TraceContext(threading.local):
    nnx_trace_stack: tp.List[MainTrace] = dataclasses.field(
        default_factory=lambda: [current_jax_trace()]
    )


TRACE_CONTEXT = TraceContext()


@contextlib.contextmanager
def nnx_trace(trace: MainTrace):
    TRACE_CONTEXT.nnx_trace_stack.append(trace)
    try:
        yield
    finally:
        TRACE_CONTEXT.nnx_trace_stack.pop()


def current_nnx_trace() -> MainTrace:
    return TRACE_CONTEXT.nnx_trace_stack[-1]


class TraceState(reprlib.Representable):
    __slots__ = ["_jax_trace", "_nnx_trace"]

    def __init__(self):
        self._jax_trace = current_jax_trace()
        self._nnx_trace = current_nnx_trace()

    @property
    def jax_trace(self):
        return self._jax_trace

    @property
    def nnx_trace(self):
        return self._nnx_trace

    def is_valid(self) -> bool:
        return (
            self._jax_trace is current_jax_trace()
            and self._nnx_trace is current_nnx_trace()
        )

    def __nnx_repr__(self):
        yield reprlib.Object(f"{type(self).__name__}")
        yield reprlib.Attr("jax_trace", self._jax_trace)
        yield reprlib.Attr("nnx_trace", self._nnx_trace)
